"""Base settings for a Django Rest Framework Project."""
import os
from distutils.util import strtobool
from t3.env import load_env
from t3.django.util import get_base_dir


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = get_base_dir()


# Get env dictionary.  This feeds from env, .env, and then vcap services
# For local dev, move .env.example to .env and play with settings
# In CI, a service.yaml is likely to be used
_env = load_env(dot_env_path=BASE_DIR)


# Create of merge INSTALLED_APPS
# print(globals().get('INSTALLED_APPS', []))
# INSTALLED_APPS = globals().get('INSTALLED_APPS', [])
# print(globals().get('INSTALLED_APPS', []))
DRF_INSTALLED_APPS = [
    'django_filters',
    'crispy_forms',
    'rest_framework',
    'drf_yasg',
    'rest_framework.authentication'
]


# Rest Framework Settings
REST_FRAMEWORK = {
    'DEFAULT_RENDERER_CLASSES': (
        'djangorestframework_camel_case.render.CamelCaseJSONRenderer',
    ),
    'DEFAULT_PARSER_CLASSES': (
        'djangorestframework_camel_case.parser.CamelCaseJSONParser',
    ),
    'DEFAULT_AUTHENTICATION_CLASSES': (),
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.LimitOffsetPagination',
    'EXCEPTION_HANDLER': 't3.drf.views.enveloped_exception_handler',
    'PAGE_SIZE': 25,
}

DOCS_TITLE = os.getenv('DOCS_TITLE', 'T3 API')
DOCS_VERSION = os.getenv('DOCS_VERSION', 'v1')

# Set up Swagger to Accept Authorization tokens for particular requests.
SWAGGER_SETTINGS = {
    'SECURITY_DEFINITIONS': {
        'Basic': {
            'type': 'basic'
        },
        'Bearer': {
            'type': 'apiKey',
            'name': 'Authorization',
            'in': 'header'
        }
    }
}

# ############################################################################
# Enables auth configurations for Punchh.  The Punchh Adapter returns JWTs that
# includes tokens that are necessary for both the `v1` api, as well as the `v2`
# api.
# To use, use the `t3.vendors.punchh.PunchhAuthentication` class
# See https://www.django-rest-framework.org/api-guide/authentication/ for more
# ############################################################################
# details on how auth w/ DRF works.
PUNCHH_AUTH = bool(strtobool(os.getenv('PUNCHH_AUTH', 'false')))
if PUNCHH_AUTH:
    # We're using the Punchh Client Secret to sign our JWTs.  It's supposed to
    # be secret and we want all JWTs to become invalid if we need to change it.
    PUNCHH_CLIENT_SECRET = os.environ['PUNCHH_CLIENT_SECRET']

    # Enable PUNCHH_AUTH_SERVER_VERIFY if you want to ping punchh's server with
    # a request, to ensure that the tokens in the JWT are valid.  This is
    # important when adding Punchh auth to services that are not either hitting
    # Punchh's servers directly, or are leveraging Punchh's SSO.
    PUNCHH_AUTH_SERVER_VERIFY = bool(strtobool(os.getenv('PUNCHH_AUTH_SERVER_VERIFY', 'false')))
    # PUNCHH_AUTH_SERVER_VERIFY = True if str(os.getenv('PUNCHH_AUTH_SERVER_VERIFY', 'false')).lower() == 'true' else False
    if PUNCHH_AUTH_SERVER_VERIFY:
        PUNCHH_BASE_URL = os.environ['PUNCHH_BASE_URL']
        PUNCHH_V2_BASE_URL = os.environ['PUNCHH_V2_BASE_URL']
        PUNCHH_CLIENT_KEY = os.environ['PUNCHH_CLIENT_KEY']



import requests
import redis

from cachecontrol import CacheControlAdapter
from cachecontrol.heuristics import ExpiresAfter
from cachecontrol.caches.redis_cache import RedisCache

from django.conf import settings
import logging
import re

log = logging.getLogger(__name__)


def log_cache_obj(message, redis_url, pattern, minutes, url):
    password_sub = re.sub('(password=)[^&\n\r]+', 'password=xxxxxxxxxx', redis_url)
    log.info({
        'message': message,
        'redis': password_sub,
        'pattern': pattern,
        'minutes': minutes,
        'url': url,
    })


def get_cache_requests(url):
    # Check settings
    if not hasattr(settings, 'EXTERNAL_CACHE'):
        return None
    external_cache = settings.EXTERNAL_CACHE
    redis_url = external_cache.get('redis')
    pattern = external_cache.get('pattern')
    minutes = external_cache.get('minutes')
    if (redis_url or pattern or minutes) is None:
        log_mess = 'Unable to use cache requests -> missing information'
        log_cache_obj(log_mess, redis_url, pattern, minutes, url)
        return None
    # Check if redis is connected
    r = redis.from_url(redis_url)
    try:
        r.info()
    except redis.ConnectionError:
        log_mess = 'Unable to use cache requests -> no redis connection'
        log_cache_obj(log_mess, redis_url, pattern, minutes, url)
        return None  # No cache connection, cannot use cache
    session = requests.session()
    # Check if url path matches (RegEx: https://regexr.com/)
    match = re.search(r'%s' % pattern, url)
    if match is None:
        log_mess = 'Unable to use cache requests -> url does not match'
        log_cache_obj(log_mess, redis_url, pattern, minutes, url)
        return None
    # Everything is connected and matches, can use cached request
    adapter = CacheControlAdapter(heuristic=ExpiresAfter(minutes=int(minutes)), cache=RedisCache(r))
    session.mount('https://', adapter)
    session.mount('http://', adapter)
    log_mess = 'Using cache requests'
    log_cache_obj(log_mess, redis_url, pattern, minutes, url)
    return session


# HTTP Get Request
def get(url, **kwargs):
    req = get_cache_requests(url)  # Use cache if available
    if req is not None:
        return req.get(url, **kwargs)
    return requests.get(url, **kwargs)


# HTTP Options Request
def options(url, **kwargs):
    return requests.options(url, **kwargs)


# HTTP Head Request
def head(url, **kwargs):
    return requests.head(url, **kwargs)


# HTTP Post Request
def post(url, **kwargs):
    return requests.post(url, **kwargs)


# HTTP Put Request
def put(url, **kwargs):
    return requests.put(url, **kwargs)


# HTTP Patch Request
def patch(url, **kwargs):
    return requests.patch(url, **kwargs)


# HTTP Delete Request
def delete(url, **kwargs):
    return requests.delete(url, **kwargs)



import signal
from t3.events.base_publisher import BasePublisher
from t3.events.util import get_task_name, get_topic_name, get_rabbit_mq_conn


class TaskPublisher(BasePublisher):
    """
    TaskPublisher is used for tasks, first consumer subscribed to the same queue_name.

    Messages are distributed in round robin fashion
    Any system can send a task to be consumed by the first available worker
    """

    def __init__(self):
        amqp_url = get_rabbit_mq_conn()
        BasePublisher.__init__(self, amqp_url)
        self.set_exchange('task_exchange')
        signal.signal(signal.SIGINT, lambda signum, frame: self.stop())
        signal.signal(signal.SIGTERM, lambda signum, frame: self.stop())

    def set_task_name(self, task_name):
        name = get_task_name(task_name)
        self.set_routing_key(name)


class TopicPublisher(BasePublisher):
    """
    TopicPublisher is used for topics, all consumers subscribed to the routing_key gets the message.

    Messages are delivered at the same time.
    Any system can send a topic to be consumed by the subscribed workers
    """

    def __init__(self):
        amqp_url = get_rabbit_mq_conn()
        BasePublisher.__init__(self, amqp_url)
        self.set_exchange('topic_exchange')
        signal.signal(signal.SIGINT, lambda signum, frame: self.stop())
        signal.signal(signal.SIGTERM, lambda signum, frame: self.stop())

    def set_topic_name(self, topic_name):
        name = get_topic_name(topic_name)
        self.set_routing_key(name)



from rest_framework import serializers


class ResourceSerializer(serializers.Serializer):
    resource_class = None

    def create(self, validated_data):
        user = None
        validated_data['request'] = self.context.get("request")

        assert self.resource_class is not None, '`resource_class` missing'
        resource = self.resource_class()  # noqa
        return resource.create(**validated_data)

    def update(self, instance, validated_data):
        validated_data['request'] = self.context.get("request")
        if self.partial:
            return instance.partial_update(**validated_data)
        return instance.update(**validated_data)

    def partial_update(self, instance, validated_data):
        validated_data['request'] = self.context.get("request")
        return instance.partial_update(**validated_data)



import sys
import traceback
from collections import OrderedDict


def json_traceback():
    """Gathers exception details, such as trace info."""
    exc_info = sys.exc_info()
    traceback_list = ''.join(traceback.format_exc()).split('\n')

    return OrderedDict([
        ('exception_class', str(exc_info[0])),
        ('exception_message', str(exc_info[1])),
        ('exception_traceback', traceback_list),
    ])


def get_response_type(code):
    if 100 <= code <= 199:
        return 'informational'
    if 200 <= code <= 299:
        return 'success'
    if 300 <= code <= 399:
        return 'redirect'
    if 400 <= code <= 499:
        return 'client_error'
    if 500 <= code <= 599:
        return 'server_error'
    return None



from rest_framework import exceptions


class RequestException(exceptions.APIException):
    def __init__(self, detail=None, code=None, status_code=None):
        super().__init__(detail, code)
        if status_code:
            self.status_code = status_code



import logging
import pika

log = logging.getLogger(__name__)


class BaseConsumer(object):
    """
    BaseConsumer will handle unexpected interactions with RabbitMQ such as channel and connection closures.

    If RabbitMQ closes the connection, it will reopen it. You should
    look at the output, as there are limited reasons why the connection may
    be closed, which usually are tied to permission related issues or
    socket timeouts.
    If the channel is closed, it will indicate a problem with one of the
    commands that were issued and that should surface in the output as well.
    """

    def __init__(self, amqp_url):
        """
        Create a new instance of the consumer class, passing in the AMQP URL used to connect to RabbitMQ.

        :param str amqp_url: The AMQP url to connect with
        """
        self._connection = None
        self._channel = None
        self._closing = False
        self._consumer_tag = None
        self._url = amqp_url

        self._queue = ''
        self._on_message_callback = None   # Use a callback to process the message.
        self._message_count = 0            # Count messages and failed messages.
        self._failed_message_count = 0

        self._exchange_name = ''
        self._exchange_type = 'direct'
        self._routing_key = ''

    def set_queue(self, routing_key, queue=''):
        self._routing_key = routing_key
        self._queue = queue

    def set_exchange(self, exchange_name, exchange_type='direct'):
        self._exchange_name = exchange_name
        self._exchange_type = exchange_type

    def add_on_message_callback(self, on_message_callback):
        self._on_message_callback = on_message_callback

    def connect(self):
        """
        Connect to RabbitMQ, returning the connection handle.

        When the connection is established, the on_connection_open method
        will be invoked by pika.
        :rtype: pika.SelectConnection
        """
        log.info('Connecting to %s', self._url)
        return pika.SelectConnection(pika.URLParameters(self._url),
                                     self.on_connection_open,
                                     self.on_connection_error,   # !! Handle error opening connection
                                     stop_ioloop_on_close=False)

    def on_connection_open(self, unused_connection):
        """
        Call by pika once the connection to RabbitMQ has been established.

        It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :type unused_connection: pika.SelectConnection
        """
        log.info('Connection opened')
        self.add_on_connection_close_callback()
        self.open_channel()

    def on_connection_error(self, unused_connection, message):
        """
        Call by pika if the connection to RabbitMQ has failed to be established.

        It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :type unused_connection: pika.SelectConnection
        """
        log.error('Connection failed, closing: %s', message)
        self.stop()

    def add_on_connection_close_callback(self):
        """Add on close callback, invoked when RabbitMQ closes the connection to the publisher unexpectedly."""
        log.info('Adding connection close callback')
        self._connection.add_on_close_callback(self.on_connection_closed)

    def on_connection_closed(self, connection, reply_code, reply_text):
        """
        Invoke by pika when the connection to RabbitMQ is closed unexpectedly.

        Since it is unexpected, we will reconnect to RabbitMQ if it disconnects.
        :param pika.connection.Connection connection: The closed connection obj
        :param int reply_code: The server provided reply_code if given
        :param str reply_text: The server provided reply_text if given
        """
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            log.warning('Connection closed, reopening in 5 seconds: (%s) %s', reply_code, reply_text)
            self._connection.add_timeout(5, self.reconnect)

    def reconnect(self):
        """
        Will be invoked by the IOLoop timer if the connection is closed.

        See the on_connection_closed method.
        """
        # This is the old connection IOLoop instance, stop its ioloop
        self._connection.ioloop.stop()

        if not self._closing:

            # Create a new connection
            self._connection = self.connect()

            # There is now a new connection, needs a new ioloop to run
            self._connection.ioloop.start()

    def open_channel(self):
        """
        Open a new channel with RabbitMQ by issuing the Channel.

        Open RPC command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.
        """
        log.info('Creating a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """
        Invoke by pika when the channel has been opened.

        The channel object is passed in so we can make use of it.
        Since the channel is now open, we'll declare the exchange to use.
        :param pika.channel.Channel channel: The channel object
        """
        log.info('Channel opened')
        self._channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange(self._exchange_name)

    def add_on_channel_close_callback(self):
        """Tells pika to call the on_channel_closed method if RabbitMQ unexpectedly closes the channel."""
        log.info('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reply_code, reply_text):
        """
        Invoke by pika when RabbitMQ unexpectedly closes the channel.

        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.
        :param pika.channel.Channel: The closed channel
        :param int reply_code: The numeric reason the channel was closed
        :param str reply_text: The text reason the channel was closed
        """
        log.warning('Channel %i was closed: (%s) %s', channel, reply_code, reply_text)
        self._connection.close()

    def setup_exchange(self, exchange_name):
        """
        Start the exchange on RabbitMQ by invoking the Exchange.Declare RPC command.

        When it is complete, the on_exchange_declareok method will be invoked by pika.
        :param str|unicode exchange_name: The name of the exchange to declare
        """
        log.info('Declaring exchange %s', exchange_name)
        self._channel.exchange_declare(self.on_exchange_declareok,
                                       exchange_name,
                                       self._exchange_type)

    def on_exchange_declareok(self, unused_frame):
        """
        Invoke by pika when RabbitMQ has finished the Exchange.Declare RPC command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame
        """
        log.info('Exchange declared')
        self.setup_queue(self._queue)

    def setup_queue(self, queue_name):
        """
        Start the queue on RabbitMQ by invoking the Queue.Declare RPC command.

        When it is complete, the on_queue_declareok method will be invoked by pika.
        :param str|unicode queue_name: The name of the queue to declare.
        """
        log.info('Declaring queue %s', queue_name)
        if len(queue_name) > 0:
            self._channel.queue_declare(self.on_queue_declareok, queue_name)
        else:
            self._channel.queue_declare(self.on_queue_declareok, exclusive=True)

    def on_queue_declareok(self, method_frame):
        """
        Invoke by pika when the Queue.Declare RPC call made in setup_queue has completed.

        In this method we will bind the queue and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.
        :param pika.frame.Method method_frame: The Queue.DeclareOk frame
        """
        log.info('Binding %s to %s with %s', self._exchange_name, self._queue, self._routing_key)
        self._channel.queue_bind(self.on_bindok, self._queue,
                                 self._exchange_name, self._routing_key)

    def on_bindok(self, unused_frame):
        """
        Invoke by pika when the Queue.Bind method has completed.

        At this point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.
        :param pika.frame.Method unused_frame: The Queue.BindOk response frame
        """
        log.info('Queue bound')
        self.start_consuming()

    def start_consuming(self):
        """
        start_consuming sets up the consumer.

        Done by first calling add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.
        """
        log.info('Issuing consumer related RPC commands')
        self.add_on_cancel_callback()
        self._consumer_tag = self._channel.basic_consume(self.on_message,
                                                         self._queue)

    def add_on_cancel_callback(self):
        """
        Add a callback that will be invoked if RabbitMQ cancels the consumer for some reason.

        If RabbitMQ does cancel the consumer, on_consumer_cancelled will be invoked by pika.
        """
        log.info('Adding consumer cancellation callback')
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        """
        Invoke by pika when RabbitMQ sends a Basic.Cancel for a consumer receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        log.info('Consumer was cancelled remotely, shutting down: %r', method_frame)
        if self._channel:
            self._channel.close()

    def on_message(self, unused_channel, basic_deliver, properties, body):
        """
        Invoke by pika when a message is delivered from RabbitMQ.

        The channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.
        :param pika.channel.Channel unused_channel: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param str|unicode body: The message body
        """
        log.info('Received message # %s from %s: %s', basic_deliver.delivery_tag, properties.app_id, body)

        acknowledge_this_message = True
        if self._on_message_callback:
            try:
                self._on_message_callback(body)
            except Exception:
                acknowledge_this_message = False

        self._message_count = self._message_count + 1
        if acknowledge_this_message:
            self.acknowledge_message(basic_deliver.delivery_tag)
        else:
            self._failed_message_count = self._failed_message_count + 1
            self.reject_message(basic_deliver.delivery_tag)
        log.info('Processed %d messages. %d failed messages', self._message_count, self._failed_message_count)

    def acknowledge_message(self, delivery_tag):
        """
        Acknowledge the message delivery from RabbitMQ by sending a Basic.Ack RPC method for the delivery tag.

        :param int delivery_tag: The delivery tag from the Basic.Deliver frame
        """
        log.info('Acknowledging message %s', delivery_tag)
        self._channel.basic_ack(delivery_tag)

    def reject_message(self, delivery_tag):
        """
        Reject the message delivery from RabbitMQ by sending a Basic.Nack RPC method for the delivery tag.

        :param int delivery_tag: The deliver tag from Basic.Deliver frame
        """
        log.info('Reject message %s', delivery_tag)
        self._channel.basic_nack(delivery_tag)

    def stop_consuming(self):
        """Tell RabbitMQ that you would like to stop consuming by sending the Basic.Cancel RPC command."""
        if self._channel:
            log.info('Sending a Basic.Cancel RPC command to RabbitMQ')
            self._channel.basic_cancel(self.on_cancelok, self._consumer_tag)

    def on_cancelok(self, unused_frame):
        """
        on_cancelok is invoked by pika when RabbitMQ acknowledges the cancellation of a consumer.

        At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.
        :param pika.frame.Method unused_frame: The Basic.CancelOk frame
        """
        log.info('RabbitMQ acknowledged the cancellation of the consumer')
        self.close_channel()

    def close_channel(self):
        """
        Call to close the channel with RabbitMQ.

        Close cleanly by issuing the Channel.Close RPC command.
        """
        log.info('Closing the channel')
        self._channel.close()

    def run(self):
        """
        Run the consumer by connecting to RabbitMQ.

        Then start the IOLoop to block and allow the SelectConnection to operate.
        """
        self._connection = self.connect()
        self._connection.ioloop.start()

    def stop(self):
        """
        Clean shutdown the connection to RabbitMQ by stopping the consumer with RabbitMQ.

        When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.
        """
        log.info('Stopping')
        log.info('Processed %d messages. %d failed messages', self._message_count, self._failed_message_count)
        self._closing = True
        self.stop_consuming()
        self._connection.ioloop.stop()
        log.info('Stopped')

    def close_connection(self):
        """close_connection method closes the connection to RabbitMQ."""
        log.info('Closing connection')
        self._connection.close()



import codecs
import hashlib
import hmac
import json
import uuid
import datetime
import logging
import re
import jwt
from requests.sessions import Session
from django.conf import settings
from rest_framework import authentication
from rest_framework.exceptions import AuthenticationFailed


logger = logging.getLogger(__name__)


class PunchhException(Exception):
    """Punchh related errors."""

    pass


class NoToken(AuthenticationFailed):
    pass


class InvalidJWT(AuthenticationFailed):
    pass


class BadClaims(AuthenticationFailed):
    pass


class PunchhSession(Session):
    """Creating our own Punchh session to perform our requests with the needed Punchh client signature header."""

    def __init__(self, base_url, secret, api_version='v1'):
        super().__init__()

        self.base_url = base_url
        self.secret = secret
        self.api_version = api_version

    def _generate_signature(self, url, r_body=None):
        uri = url.split(self.base_url)[1]
        encryption_type = hashlib.sha1
        if self.api_version == 'v2':
            encryption_type = hashlib.sha256

        if r_body:
            payload = uri + r_body
        else:
            payload = uri

        # Handling of hmac library for encryption of the header x-pch signature needed by Punchh
        new_hash = hmac.new(
            bytes(self.secret.encode('utf-8')),
            bytes(payload.encode('utf-8')),
            encryption_type).digest()

        signature = codecs.encode(new_hash, 'hex_codec')

        return signature

    def prepare_request(self, request):
        if request.data:
            request.data = json.dumps(request.data, separators=(',', ':'))
            sig = self._generate_signature(request.url, request.data)
        elif request.json:
            # TODO find out about why the json returns an invalid signature
            request.json = json.dumps(request.json, separators=(',', ':'))
            # sig = self._generate_signature(request.url, request.json)

        request.headers['Content-Type'] = 'application/json'
        request.headers['Accept'] = 'application/json'
        request.headers['x-pch-digest'] = sig

        return super().prepare_request(request)


def get_token(auth: dict, api_version: str = 'v1') -> str:
    """Get the appropriate token from an auth object."""
    # api token can only be 'v1' or 'v2'
    if api_version not in ['v1', 'v2']:
        raise PunchhException(f'Invalid api_version: {api_version}')

    try:
        return auth[f'{api_version}_token']
    except KeyError:
        raise PunchhException(f"Invalid Auth Token {api_version}.")


def create_jwt(data: dict, secret: str) -> str:
    """Create JWT that is compatible with `t3.drf.authentication.PunchhAuthentication`."""
    if not all([x in data for x in ['user_id', 'authentication_token', 'access_token']]):
        raise PunchhException('Invalid data used to generate token')

    claims = {
        'jti': str(uuid.uuid4()),
        'sub': data.get('user_id', None),
        'iat': datetime.datetime.utcnow(),
        'v1_token': data.get('authentication_token', None),
        'v2_token': data.get('access_token', None),
    }

    if not claims['sub'] or not claims['v1_token']:
        raise PunchhException('Data used to create JWT is incomplete.')

    if not claims['v2_token']:
        raise PunchhException('Data used to create JWT is incomplete.  Ensure that Punchh has turned on feature that '
                              'makes the `v1` login/register  calls return both the `authentication_token` (v1), as '
                              'well as the `authentication_token` (v2)')

    token = jwt.encode(
        claims,
        secret,
        algorithm='HS256',
    )

    return token.decode("utf-8")


class PunchhAuthentication(authentication.BaseAuthentication):
    """Django Rest Framework authentication class."""

    def authenticate(self, request):
        jwt_token = re.sub(f'^Bearer ', '', request.META.get('HTTP_AUTHORIZATION', ''))


        if not jwt_token:
            raise NoToken('No token in request header')

        try:
            payload = jwt.decode(jwt_token, settings.PUNCHH_CLIENT_SECRET, algorithms='HS256')
        except:
            raise InvalidJWT('Invalid JWT')

        # Verify that the token has the proper claims
        if not all([x in payload for x in ['jti', 'sub', 'iat', 'v1_token', 'v2_token']]):
            raise BadClaims('Invalid Punchh JWT used')

        user = payload['sub']
        v2_token = payload['v2_token']

        # If we want to verify that the auth tokens inside the JWT are valid,
        # then perform a retrieve user action.  We want to do this when we're
        # hitting endpoints that don't rely on Punchh's authentication system.
        # Doing this adds a round-trip request to Punchh's servers to EVERY call
        if settings.PUNCHH_AUTH_SERVER_VERIFY:
            # Make session for sending get request to punchh
            session = PunchhSession(
                settings.PUNCHH_V2_BASE_URL,
                settings.PUNCHH_CLIENT_SECRET,
                'v2',
            )

            # Send the request
            response = session.get(
                f'{settings.PUNCHH_V2_BASE_URL}/api2/mobile/users/{user}',
                data={'client': settings.PUNCHH_CLIENT_KEY},
                headers={'Authorization': f'Bearer {v2_token}'}
            )

            if response.status_code != 200:
                raise AuthenticationFailed('Punchh servers rejected token')

        return (user, payload)
